//
// Copyright (c) 2019 Ted Unangst <tedu@tedunangst.com>
//
// Permission to use, copy, modify, and distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
// ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
// OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

package main

import (
	"fmt"
	"io/ioutil"
	"log"
	"net"
	"os"
	"sort"
	"strconv"

	"github.com/gdamore/tcell"
	"github.com/rivo/tview"
)

func adminscreen() {
	log.SetOutput(ioutil.Discard)
	allrepos := loadrepos()

	var deleted []*RepoRow

	app := tview.NewApplication()
	var maindriver func(event *tcell.EventKey) *tcell.EventKey

	table := tview.NewTable().SetFixed(1, 0).SetSelectable(true, false).
		SetSelectedStyle(tcell.ColorBlack, tcell.ColorGreen, 0)

	mainframe := tview.NewFrame(table)
	mainframe.AddText(tview.Escape("humungus - [a] add [d] delete [e] edit [s] save and quit [^C] abort"),
		true, 0, tcell.ColorGreen)
	mainframe.SetBorders(1, 0, 1, 0, 4, 0)

	pad := func(s string) string {
		return fmt.Sprintf("  %-20s", s)
	}

	dupecell := func(base *tview.TableCell) *tview.TableCell {
		rv := new(tview.TableCell)
		*rv = *base
		return rv
	}

	showtable := func() {
		table.Clear()
		sort.Slice(allrepos, func(i, j int) bool {
			return allrepos[i].Order < allrepos[j].Order
		})

		row := 0
		{
			col := 0
			headcell := tview.TableCell{
				Color:         tcell.ColorWhite,
				NotSelectable: true,
			}
			cell := dupecell(&headcell)
			cell.Text = "order"
			table.SetCell(row, col, cell)
			col++
			cell = dupecell(&headcell)
			cell.Text = "  name"
			table.SetCell(row, col, cell)
			col++
			cell = dupecell(&headcell)
			cell.Text = "description"
			table.SetCell(row, col, cell)
			col++

			row++
		}

		for _, repo := range allrepos {
			col := 0
			headcell := tview.TableCell{
				Color: tcell.ColorWhite,
			}
			cell := dupecell(&headcell)
			cell.Text = fmt.Sprintf("%d", repo.Order)
			cell.Align = tview.AlignRight
			table.SetCell(row, col, cell)
			col++
			cell = dupecell(&headcell)
			cell.Text = pad(repo.Name)
			table.SetCell(row, col, cell)
			col++
			cell = dupecell(&headcell)
			cell.Text = repo.Description
			table.SetCell(row, col, cell)
			col++

			row++
		}

		app.SetInputCapture(maindriver)
		app.SetRoot(mainframe, true)
	}

	arrowadapter := func(event *tcell.EventKey) *tcell.EventKey {
		switch event.Key() {
		case tcell.KeyDown:
			return tcell.NewEventKey(tcell.KeyTab, '\t', tcell.ModNone)
		case tcell.KeyUp:
			return tcell.NewEventKey(tcell.KeyBacktab, '\t', tcell.ModNone)
		}
		return event
	}

	editform := tview.NewForm()
	namebox := tview.NewInputField().SetLabel("name").SetFieldWidth(20)
	descbox := tview.NewInputField().SetLabel("description").SetFieldWidth(60)
	orderbox := tview.NewInputField().SetLabel("order").SetFieldWidth(10)
	editform.AddButton("save", nil)
	editform.AddButton("cancel", nil)
	savebutton := editform.GetButton(0)
	editform.SetFieldTextColor(tcell.ColorBlack)
	editform.SetFieldBackgroundColor(tcell.ColorGreen)
	editform.SetLabelColor(tcell.ColorWhite)
	editform.SetButtonTextColor(tcell.ColorGreen)
	editform.SetButtonBackgroundColor(tcell.ColorBlack)
	editform.GetButton(1).SetSelectedFunc(showtable)
	editform.SetCancelFunc(showtable)

	hadchanges := false

	showform := func() {
		editform.Clear(false)
		editform.AddFormItem(namebox)
		editform.AddFormItem(descbox)
		editform.AddFormItem(orderbox)
		app.SetInputCapture(arrowadapter)
		app.SetRoot(editform, true)
	}

	editrepo := func(r *RepoRow) {
		namebox.SetText(r.Name)
		descbox.SetText(r.Description)
		orderbox.SetText(fmt.Sprintf("%d", r.Order))
		savebutton.SetSelectedFunc(func() {
			r.Name = namebox.GetText()
			r.Description = descbox.GetText()
			r.Order, _ = strconv.Atoi(orderbox.GetText())
			hadchanges = true
			showtable()
		})
		showform()
	}

	addrepo := func() {
		namebox.SetText("")
		descbox.SetText("")
		orderbox.SetText("10")
		savebutton.SetSelectedFunc(func() {
			r := new(RepoRow)
			r.Name = namebox.GetText()
			r.Description = descbox.GetText()
			r.Order, _ = strconv.Atoi(orderbox.GetText())
			allrepos = append(allrepos, r)
			hadchanges = true
			showtable()
		})
		showform()
	}
	saveandquit := func() {
		app.Stop()
		if !hadchanges {
			return
		}
		db := getdbconn()
		defer putdbconn(db)

		tx, err := db.Conn.Begin()
		if err != nil {
			fmt.Printf("error starting tx: %s\n", err)
			return
		}
		for _, r := range allrepos {
			if r.ID == 0 {
				_, err = tx.Exec("insert into repos (name, description, orderid) values (?, ?, ?)",
					r.Name, r.Description, r.Order)
			} else {
				_, err = tx.Exec("update repos set name = ?, description = ?, orderid = ? where repoid = ?",
					r.Name, r.Description, r.Order, r.ID)
			}
			if err != nil {
				fmt.Printf("error saving row: %s\n", err)
				return
			}
		}
		for _, r := range deleted {
			_, err := tx.Exec("delete from repos where repoid = ?", r.ID)
			if err != nil {
				fmt.Printf("error deleting row: %s\n", err)
				return
			}
		}
		err = tx.Commit()
		if err != nil {
			fmt.Printf("error committing tx: %s\n", err)
			return
		}
		sendcmd("init")
	}
	maindriver = func(event *tcell.EventKey) *tcell.EventKey {
		switch event.Rune() {
		case 'a':
			addrepo()
		case 'd':
			hadchanges = true
			row, _ := table.GetSelection()
			table.RemoveRow(row)
			if row == table.GetRowCount() {
				table.Select(row-1, 0)
			}
			row--
			if row >= 0 && row < len(allrepos) {
				r := allrepos[row]
				if row == len(allrepos)-1 {
					allrepos = allrepos[:row]
				} else {
					allrepos = append(allrepos[:row], allrepos[row+1:]...)
				}
				deleted = append(deleted, r)
			}
		case 'e':
			row, _ := table.GetSelection()
			row--
			if row >= 0 && row < len(allrepos) {
				r := allrepos[row]
				editrepo(r)
			}
		case 's':
			saveandquit()
			return nil
		}
		return event
	}

	showtable()
	app.Run()
}

func sendcmd(cmd string) error {
	sockname := "humungus.sock"
	sock, err := net.Dial("unix", sockname)
	if err != nil {
		return err
	}
	sock.Write([]byte(cmd))
	sock.Close()
	return nil
}

func controlcommand(args []string) {
	if len(args) == 0 {
		fmt.Printf("missing command\n")
		os.Exit(1)
	}
	switch args[0] {
	case "stop":
		err := sendcmd("stop")
		if err != nil {
			fmt.Printf("could not connect: %s\n", err)
		}
	default:
		fmt.Printf("unknown command\n")
		os.Exit(1)
	}
}
