//
// Copyright (c) 2018 Ted Unangst <tedu@tedunangst.com>
//
// Permission to use, copy, modify, and distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
// ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
// OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

package main

import (
	"math/rand"
	"sync"
	"time"
)

func init() {
	rand.Seed(time.Now().Unix())
}

// A cache
type Cache struct {
	mtx     sync.Mutex
	cache   map[string][]byte
	size    int
	maxsize int
}

// Create a new cache with the specified maximum size
func NewCache(maxsize int) *Cache {
	if maxsize <= 0 {
		panic("invalid cache size")
	}
	return &Cache{
		cache:   make(map[string][]byte),
		maxsize: maxsize,
	}
}

// Get an item
func (c *Cache) Get(key string) []byte {
	c.mtx.Lock()
	defer c.mtx.Unlock()
	if c.cache == nil {
		return nil
	}
	data := c.cache[key]
	return data
}

// Put (add) an item
func (c *Cache) Put(key string, data []byte) {
	if len(data) > c.maxsize/2 {
		// too big, don't bother
		return
	}
	c.mtx.Lock()
	defer c.mtx.Unlock()
	if c.cache == nil {
		return
	}
	c.size += len(data)
	for c.size > c.maxsize {
		for k, v := range c.cache {
			if rand.Intn(100) < 20 {
				c.size -= len(v)
				delete(c.cache, k)
			}
		}
	}
	c.cache[key] = data
}

// Flush
func (c *Cache) Flush() {
	c.mtx.Lock()
	defer c.mtx.Unlock()
	c.cache = make(map[string][]byte)
	c.size = 0
}

// Quit (and clear)
func (c *Cache) Quit() {
	c.mtx.Lock()
	defer c.mtx.Unlock()
	c.cache = nil
}
