//
// Copyright (c) 2018 Ted Unangst <tedu@tedunangst.com>
//
// Permission to use, copy, modify, and distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
// ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
// OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

package main

import (
	"database/sql"
	"io"
	"log"
	"os"
	"os/exec"
	"sync"

	"humungus.tedunangst.com/r/gerc"
	_ "humungus.tedunangst.com/r/go-sqlite3"
)

// A common interface for connections
type CommonConn interface {
	Finish()
	ShouldFinish() bool
}

// Read directly from the repo
type GercConn struct {
	Name string
	Repo *gerc.Repo
}

// A connection to an hg repo
// Comes in two flavors, wire and command
type RepoConn struct {
	Name      string
	CmdServer bool
	Wpipe     io.WriteCloser
	Rpipe     io.ReadCloser
	Cmd       *exec.Cmd
	cache     *Cache
}

// Caches for hg connections
var cachemtx sync.Mutex
var wirecaches = make(map[string]*Cache)

func newRepoConn(reponame string, cmdserver bool) *RepoConn {
	var conntype string
	var hgargs []string
	if cmdserver {
		conntype = "cmdserver"
		hgargs = []string{"-R", reponame, "serve", "--cmdserver", "pipe"}
	} else {
		conntype = "wire"
		hgargs = []string{"-R", reponame, "serve", "--stdio"}
	}
	log.Printf("starting %s server for %s", conntype, reponame)
	cmd := exec.Command("hg", hgargs...)
	cmd.Dir = "repos"
	wpipe, _ := cmd.StdinPipe()
	rpipe, _ := cmd.StdoutPipe()
	err := cmd.Start()
	if err != nil {
		log.Printf("error running cmd: %s", err)
		return nil
	}

	conn := &RepoConn{
		Name:      reponame,
		CmdServer: cmdserver,
		Wpipe:     wpipe,
		Rpipe:     rpipe,
		Cmd:       cmd,
	}

	if !cmdserver {
		cachemtx.Lock()
		cache := wirecaches[reponame]
		if cache == nil {
			cache = NewCache(10 * 1024 * 1024)
			wirecaches[reponame] = cache
		}
		cachemtx.Unlock()
		conn.cache = cache
	}

	if cmdserver {
		success := false
		defer func() {
			if !success {
				conn.Finish()
			}
		}()
		readdata(rpipe)
		success = true
	}
	return conn
}

func creategercconn(reponame string) CommonConn {
	repo, _ := gerc.Open("repos/" + reponame)
	conn := &GercConn{
		Name: reponame,
		Repo: repo,
	}
	return conn
}

func (conn *GercConn) Finish() {
	conn.Repo.Close()
}

func (conn *GercConn) ShouldFinish() bool {
	return false
}

func createcmdconn(reponame string) CommonConn {
	return newRepoConn(reponame, true)
}

func createwireconn(reponame string) CommonConn {
	return newRepoConn(reponame, false)
}

func (conn *RepoConn) Finish() {
	if conn.Wpipe != nil {
		conn.Wpipe.Close()
	}
	conn.Rpipe.Close()
	conn.Cmd.Wait()
}

func (conn *RepoConn) ShouldFinish() bool {
	return conn.Wpipe == nil
}

// A connection to the database
type DbConn struct {
	Conn *sql.DB
}

func createdbconn(string) CommonConn {
	dbname := "humungus.db"
	log.Printf("opening database: %s", dbname)
	_, err := os.Stat(dbname)
	if err != nil {
		log.Fatalf("unable to open database: %s", err)
	}
	db, err := sql.Open("sqlite3", dbname)
	if err != nil {
		log.Fatalf("unable to open database: %s", err)
	}
	dconn := &DbConn{
		Conn: db,
	}
	return dconn
}

func (dconn *DbConn) Finish() {
	dconn.Conn.Close()
}

func (dconn *DbConn) ShouldFinish() bool {
	return false
}

// A pool of connections to the same backend
type ConnectionPool struct {
	Name     string
	Creator  func(string) CommonConn
	conns    []CommonConn
	numconns int
	maxconns int
	dead     bool
	mtx      sync.Mutex
	cond     *sync.Cond
}

func newConnectionPool(name string, maxconns int, creator func(string) CommonConn) *ConnectionPool {
	pool := &ConnectionPool{
		Name:     name,
		Creator:  creator,
		maxconns: maxconns,
	}
	pool.cond = sync.NewCond(&pool.mtx)
	return pool
}

func (pool *ConnectionPool) Get() CommonConn {
	pool.mtx.Lock()
	locked := true
	defer func() {
		if locked {
			pool.mtx.Unlock()
		}
	}()

	slen := len(pool.conns)
	for slen == 0 {
		if pool.numconns == pool.maxconns {
			log.Printf("waiting for a conn to %s", pool.Name)
			pool.cond.Wait()
			slen = len(pool.conns)
		} else {
			failed := 1
			pool.numconns += 1
			locked = false
			pool.mtx.Unlock()
			var rv CommonConn
			defer func() { pool.numconns -= failed }()
			rv = pool.Creator(pool.Name)
			failed = 0
			return rv
		}
	}
	conn := pool.conns[slen-1]
	pool.conns = pool.conns[:slen-1]
	return conn
}

func (pool *ConnectionPool) Put(conn CommonConn) {
	pool.mtx.Lock()
	locked := true
	defer func() {
		if locked {
			pool.mtx.Unlock()
		}
	}()

	if conn.ShouldFinish() || pool.dead {
		pool.numconns -= 1
		locked = false
		pool.mtx.Unlock()
		pool.cond.Signal()
		conn.Finish()
		return
	}

	pool.conns = append(pool.conns, conn)
	pool.cond.Signal()
}

// Pools for each connection type for each repo
var repopoolmtx sync.Mutex
var repogercpool = make(map[string]*ConnectionPool)
var repocmdpool = make(map[string]*ConnectionPool)
var repowirepool = make(map[string]*ConnectionPool)

var gercconnmap = make(map[*gerc.Repo]*GercConn)

// Unwrap connection for callers convenience
func getgerc(reponame string) *gerc.Repo {
	repopoolmtx.Lock()
	pool := repogercpool[reponame]
	repopoolmtx.Unlock()
	if pool == nil {
		return nil
	}
	conn := pool.Get().(*GercConn)
	repopoolmtx.Lock()
	gercconnmap[conn.Repo] = conn
	repopoolmtx.Unlock()
	return conn.Repo
}

func putgerc(repo *gerc.Repo) {
	repopoolmtx.Lock()
	conn := gercconnmap[repo]
	delete(gercconnmap, repo)
	reponame := conn.Name
	pool := repogercpool[reponame]
	repopoolmtx.Unlock()
	pool.Put(conn)
}

func getcmdconn(reponame string) *RepoConn {
	repopoolmtx.Lock()
	pool := repocmdpool[reponame]
	repopoolmtx.Unlock()
	if pool == nil {
		return nil
	}
	return pool.Get().(*RepoConn)
}

func getwireconn(reponame string) *RepoConn {
	repopoolmtx.Lock()
	pool := repowirepool[reponame]
	repopoolmtx.Unlock()
	if pool == nil {
		return nil
	}
	return pool.Get().(*RepoConn)
}

func putbackconn(conn *RepoConn) {
	var pool *ConnectionPool
	reponame := conn.Name
	repopoolmtx.Lock()
	if conn.CmdServer {
		pool = repocmdpool[reponame]
	} else {
		pool = repowirepool[reponame]
	}
	repopoolmtx.Unlock()
	pool.Put(conn)
}

// Database connection pool
var dbpool = newConnectionPool("dbconn", 2, createdbconn)

func getdbconn() *DbConn {
	return dbpool.Get().(*DbConn)
}

func putdbconn(dconn *DbConn) {
	dbpool.Put(dconn)
}
