//
// Copyright (c) 2018 Ted Unangst <tedu@tedunangst.com>
//
// Permission to use, copy, modify, and distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
// ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
// OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

package main

import (
	"bufio"
	"bytes"
	"crypto/sha512"
	"fmt"
	"html"
	"html/template"
	"io"
	"io/ioutil"
	"log"
	"net"
	"net/http"
	"os"
	"sort"
	"strings"
	"sync"
	"time"

	"github.com/gorilla/mux"
	"humungus.tedunangst.com/r/gerc"
)

var serverName string

var hlcache *Cache

var htmlfragCodeline = []byte("\n<span class=codeline></span>")

func highlight(code []byte, filename string) []byte {
	hasher := sha512.New()
	io.WriteString(hasher, filename)
	io.WriteString(hasher, "\x00")
	hasher.Write(code)
	key := string(hasher.Sum(nil))
	hh := hlcache.Get(key)
	if hh != nil {
		return hh
	}

	var buf bytes.Buffer
	highlightcode(code, filename, &buf, true)

	hh = buf.Bytes()
	if len(hh) > 0 && hh[len(hh)-1] == '\n' {
		hh = hh[:len(hh)-1]
	}
	if filename != "hg.diff" && len(hh) > 0 {
		hh = append(htmlfragCodeline[1:],
			bytes.Replace(hh, []byte("\n"), htmlfragCodeline, -1)...)
	}

	hlcache.Put(key, hh)
	return hh
}

func getInfo(r *http.Request) map[string]interface{} {
	templinfo := make(map[string]interface{})
	templinfo["StyleParam"] = getstyleparam()
	return templinfo
}

func filelisting(repo *gerc.Repo, reponame string, path string) []string {
	files, _ := repo.GetFiles(gerc.FilesArgs{Filenames: []string{path}})
	dirname := ""
	prefix := path
	var lines []string
	var dirlines []string
	for _, file := range files {
		line := file.Name
		if line == ".hgtags" {
			continue
		}
		line = strings.TrimPrefix(line, prefix)
		if idx := strings.IndexByte(line, '/'); idx != -1 {
			if line[0:idx+1] == dirname {
				continue
			}
			dirname = line[0 : idx+1]
			dirlines = append(dirlines, path+dirname)
			continue
		}
		lines = append(lines, path+line)
	}
	return append(dirlines, lines...)
}

type Change struct {
	Changeset string
	Parent    string
	Branch    string
	Tag       string
	User      string
	Date      time.Time
	Summary   string
	Message   []string
}

func changelisting(conn *gerc.Repo, changeset string, filename string, limit int) []*Change {
	if changeset == "" {
		changeset = "tip"
	}
	if strings.IndexByte(changeset, ':') == -1 {
		changeset += ":0"
	}
	if limit > 0 {
		changeset = fmt.Sprintf("limit(%s, %d)", changeset, limit)
	}

	gchanges, err := conn.GetChanges(gerc.ChangesArgs{Revisions: changeset, Filename: filename})
	if err != nil {
		log.Printf("error getting changes: %s\n", err)
		return nil
	}
	var changes []*Change
	for _, g := range gchanges {
		changes = append(changes, &Change{
			Changeset: fmt.Sprintf("%.6x", g.NodeID),
			//Parent:    parent,
			//Branch:    branch,
			//Tag:       tag,
			User:    g.User,
			Date:    g.Date,
			Summary: g.Summary,
			Message: g.Message,
		})
	}
	return changes
}

func recentchangelisting(conn *gerc.Repo) []*Change {
	changes := changelisting(conn, "date(yesterday):", "", 0)
	if len(changes) < 10 {
		changes = changelisting(conn, "", "", 10)
	} else {
		// reverse it
		for i, j := 0, len(changes)-1; i < j; i, j = i+1, j-1 {
			changes[i], changes[j] = changes[j], changes[i]
		}
	}
	return changes
}

func getreadme(conn *gerc.Repo) string {
	readmenames := []string{"README", "README.txt", "README.md"}
	files, err := conn.GetFiles(gerc.FilesArgs{Filenames: readmenames})
	if err != nil || len(files) == 0 {
		return ""
	}
	data, err := conn.GetFileData(gerc.FileDataArgs{Filename: files[0].Name})
	if err != nil || len(data) == 0 {
		return ""
	}
	return string(data)
}

var templates *Template

func showrepo(w http.ResponseWriter, r *http.Request) {
	if strings.HasPrefix(r.Header.Get("User-Agent"), "mercurial/proto-1.0") {
		hgpassthru(w, r)
		return
	}
	reponame := mux.Vars(r)["reponame"]
	conn := getgerc(reponame)
	if conn == nil {
		http.NotFound(w, r)
		return
	}
	defer putgerc(conn)

	changes := recentchangelisting(conn)
	if changes == nil {
		http.Error(w, "sadness", 503)
		return
	}

	w.Header().Set("Cache-Control", "max-age=30")

	templinfo := getInfo(r)
	templinfo["ServerName"] = serverName
	templinfo["Repo"] = reponame
	templinfo["Readme"] = getreadme(conn)
	templinfo["Changes"] = changes
	templinfo["HasDownloads"] = hasdownloads(reponame)
	err := templates.ExecuteTemplate(w, "overview.html", templinfo)
	if err != nil {
		log.Printf("template execution: %s", err)
	}
}

func showfiles(w http.ResponseWriter, r *http.Request) {
	reponame := mux.Vars(r)["reponame"]
	filename := mux.Vars(r)["filename"]
	changeset := mux.Vars(r)["changeset"]
	if changeset == "" {
		changeset = "tip"
	}
	repo := getgerc(reponame)
	if repo == nil {
		http.NotFound(w, r)
		return
	}
	defer putgerc(repo)

	path := strings.Split(filename, "/")
	basename := path[len(path)-1]
	path = path[:len(path)-1]

	w.Header().Set("Cache-Control", "max-age=30")

	templinfo := getInfo(r)
	templinfo["Repo"] = reponame
	templinfo["Changeset"] = changeset
	templinfo["Path"] = path
	templinfo["Basename"] = basename
	templinfo["Files"] = filelisting(repo, reponame, filename)
	templinfo["HasDownloads"] = hasdownloads(reponame)
	err := templates.ExecuteTemplate(w, "files.html", templinfo)
	if err != nil {
		log.Printf("template execution: %s", err)
	}
}

var dlfilescache = make(map[string][]os.FileInfo)
var dlfilescachemtx sync.Mutex
var dlfilescachetime time.Time

func finddownloads(reponame string) []os.FileInfo {
	dlfilescachemtx.Lock()
	defer dlfilescachemtx.Unlock()
	if time.Now().After(dlfilescachetime.Add(1 * time.Minute)) {
		dlfilescache = make(map[string][]os.FileInfo)
		dlfilescachetime = time.Now()
	}

	files := dlfilescache[reponame]
	if files != nil {
		return files
	}

	dirname := "downloads/" + reponame
	dir, err := os.Open(dirname)
	if err == nil {
		defer dir.Close()
		names, _ := dir.Readdirnames(0)
		for _, name := range names {
			fi, err := os.Stat(dirname + "/" + name)
			if err == nil {
				files = append(files, fi)
			}
		}
	}
	sort.Slice(files, func(i, j int) bool { return files[i].ModTime().After(files[j].ModTime()) })

	dlfilescache[reponame] = files
	return files
}

func hasdownloads(reponame string) bool {
	return len(finddownloads(reponame)) > 0
}

func showdownloads(w http.ResponseWriter, r *http.Request) {
	reponame := mux.Vars(r)["reponame"]
	conn := getcmdconn(reponame)
	if conn == nil {
		http.NotFound(w, r)
		return
	}
	defer putbackconn(conn)

	files := finddownloads(reponame)

	w.Header().Set("Cache-Control", "max-age=30")

	templinfo := getInfo(r)
	templinfo["Repo"] = reponame
	templinfo["Files"] = files
	err := templates.ExecuteTemplate(w, "downloads.html", templinfo)
	if err != nil {
		log.Printf("template execution: %s", err)
	}
}

func downloadfile(w http.ResponseWriter, r *http.Request) {
	reponame := mux.Vars(r)["reponame"]
	filename := mux.Vars(r)["filename"]
	conn := getcmdconn(reponame)
	if conn == nil {
		http.NotFound(w, r)
		return
	}
	defer putbackconn(conn)

	log.Printf("download of file %s", filename)

	fd, err := os.Open("downloads/" + reponame + "/" + filename)
	if err != nil {
		http.NotFound(w, r)
		return
	}
	defer fd.Close()

	w.Header().Set("Cache-Control", "max-age=432000")
	w.Header().Set("Content-Disposition", fmt.Sprintf(`attachment; filename="%s"`, filename))

	io.Copy(w, fd)
}

func showannotate(w http.ResponseWriter, r *http.Request) {
	reponame := mux.Vars(r)["reponame"]
	filename := mux.Vars(r)["filename"]
	changeset := mux.Vars(r)["changeset"]
	conn := getgerc(reponame)
	if conn == nil {
		http.NotFound(w, r)
		return
	}
	defer putgerc(conn)

	rawlines, err := conn.Annotate(gerc.AnnotateArgs{Filename: filename, Revisions: changeset})
	if err != nil {
		log.Printf("error annotating: %s", err)
		http.Error(w, "sadness", 503)
		return
	}
	numlines := len(rawlines)
	clines := make([][]byte, numlines)
	codelines := make([][]byte, numlines)
	for i, line := range rawlines {
		clines[i] = []byte(line.Annotation)
		codelines[i] = []byte(line.Line)
	}
	codelines = append(codelines, []byte(""))
	codelines = bytes.Split(highlight(bytes.Join(codelines, []byte("\n")), filename), []byte("\n"))
	for i := 0; i < len(clines); i++ {
		codelines[i] = []byte(fmt.Sprintf(
			`<a href="/r/%s/v/%s">%s</a>:	%s`,
			reponame, clines[i], clines[i], codelines[i]))
	}

	path := strings.Split(filename, "/")
	basename := path[len(path)-1]
	path = path[:len(path)-1]

	w.Header().Set("Cache-Control", "max-age=30")

	templinfo := getInfo(r)
	templinfo["Repo"] = reponame
	templinfo["Changeset"] = changeset
	templinfo["Filename"] = filename
	templinfo["Path"] = path
	templinfo["Basename"] = basename
	templinfo["Contents"] = template.HTML(bytes.Join(codelines, []byte("\n")))
	templinfo["HasDownloads"] = hasdownloads(reponame)
	templinfo["IsAnnotated"] = true
	err = templates.ExecuteTemplate(w, "file.html", templinfo)
	if err != nil {
		log.Printf("template execution: %s", err)
	}
}

func showfileraw(w http.ResponseWriter, r *http.Request) {
	reponame := mux.Vars(r)["reponame"]
	filename := mux.Vars(r)["filename"]
	changeset := mux.Vars(r)["changeset"]
	conn := getgerc(reponame)
	if conn == nil {
		http.NotFound(w, r)
		return
	}
	defer putgerc(conn)

	w.Header().Set("Cache-Control", "max-age=30")

	data, err := conn.GetFileData(gerc.FileDataArgs{Filename: filename, Revision: changeset})
	if err != nil {
		log.Printf("error getting file data: %s\n", err)
		http.Error(w, "sadness", 503)
		return
	}

	w.Write(data)
}

func showfile(w http.ResponseWriter, r *http.Request) {
	reponame := mux.Vars(r)["reponame"]
	filename := mux.Vars(r)["filename"]
	changeset := mux.Vars(r)["changeset"]
	if filename == "" || strings.HasSuffix(filename, "/") {
		showfiles(w, r)
		return
	}
	conn := getgerc(reponame)
	if conn == nil {
		http.NotFound(w, r)
		return
	}
	defer putgerc(conn)

	data, err := conn.GetFileData(gerc.FileDataArgs{Filename: filename, Revision: changeset})
	if err != nil {
		log.Printf("error getting file data: %s\n", err)
		http.Error(w, "sadness", 503)
		return
	}

	path := strings.Split(filename, "/")
	basename := path[len(path)-1]
	path = path[:len(path)-1]

	w.Header().Set("Cache-Control", "max-age=30")

	templinfo := getInfo(r)
	templinfo["Repo"] = reponame
	templinfo["Changeset"] = changeset
	templinfo["Filename"] = filename
	templinfo["Path"] = path
	templinfo["Basename"] = basename
	templinfo["Contents"] = template.HTML(highlight(data, filename))
	templinfo["HasDownloads"] = hasdownloads(reponame)
	err = templates.ExecuteTemplate(w, "file.html", templinfo)
	if err != nil {
		log.Printf("template execution: %s", err)
	}
}

func showchanges(w http.ResponseWriter, r *http.Request) {
	reponame := mux.Vars(r)["reponame"]
	changeset := mux.Vars(r)["changeset"]
	filename := mux.Vars(r)["filename"]
	conn := getgerc(reponame)
	if conn == nil {
		http.NotFound(w, r)
		return
	}
	defer putgerc(conn)

	path := strings.Split(filename, "/")
	basename := path[len(path)-1]
	path = path[:len(path)-1]

	changes := changelisting(conn, changeset, filename, 0)
	if changes == nil {
		http.Error(w, "sadness", 503)
		return
	}

	w.Header().Set("Cache-Control", "max-age=30")

	templinfo := getInfo(r)
	templinfo["Repo"] = reponame
	templinfo["Changeset"] = changeset
	templinfo["Filename"] = filename
	templinfo["Path"] = path
	templinfo["Basename"] = basename
	templinfo["Changes"] = changes
	templinfo["HasDownloads"] = hasdownloads(reponame)
	err := templates.ExecuteTemplate(w, "changes.html", templinfo)
	if err != nil {
		log.Printf("template execution: %s", err)
	}
}

func showreporss(w http.ResponseWriter, r *http.Request) {
	reponame := mux.Vars(r)["reponame"]
	conn := getgerc(reponame)
	if conn == nil {
		http.NotFound(w, r)
		return
	}
	defer putgerc(conn)

	changes := recentchangelisting(conn)
	if changes == nil {
		http.Error(w, "sadness", 503)
		return
	}

	feed := RssFeed{
		Title:       reponame + " changes",
		Description: "most recent changes for " + reponame,
		Link:        fmt.Sprintf("https://%s/r/%s", serverName, reponame),
	}

	for _, c := range changes {
		desc := fmt.Sprintf("<p>Changeset %s by %s:\n<p>%s\n",
			c.Changeset, c.User, strings.Replace(html.EscapeString(strings.Join(c.Message, "\n")), "\n", "<br>\n", -1))
		feed.Items = append(feed.Items, &RssItem{
			Title:       c.Summary,
			Description: desc,
			Link:        fmt.Sprintf("https://%s/r/%s/v/%s", serverName, reponame, c.Changeset),
			PubDate:     c.Date.Format(time.RFC1123),
		})
	}

	w.Header().Set("Cache-Control", "max-age=30")

	err := feed.Write(w)
	if err != nil {
		log.Printf("rss error: %s", err)
	}
}

func showchangeset(w http.ResponseWriter, r *http.Request) {
	reponame := mux.Vars(r)["reponame"]
	changeset := mux.Vars(r)["changeset"]
	conn := getgerc(reponame)
	if conn == nil {
		http.NotFound(w, r)
		return
	}
	defer putgerc(conn)

	changes, err := conn.GetChanges(gerc.ChangesArgs{Revisions: changeset, WithDiff: true})
	if err != nil {
		log.Printf("error getting changeset: %s\n", err)
		http.Error(w, "sadness", 503)
		return
	}
	var data strings.Builder
	if len(changes) == 1 {
		change := changes[0]
		change.Print(&data, true)
	}

	w.Header().Set("Cache-Control", "max-age=30")

	templinfo := getInfo(r)
	templinfo["Repo"] = reponame
	templinfo["Changeset"] = changeset
	templinfo["Contents"] = template.HTML(highlight([]byte(data.String()), "hg.diff"))
	templinfo["HasDownloads"] = hasdownloads(reponame)
	err = templates.ExecuteTemplate(w, "changeset.html", templinfo)
	if err != nil {
		log.Printf("template execution: %s", err)
	}
}
func rootpage(w http.ResponseWriter, r *http.Request) {
	allrepos := loadrepos()

	w.Header().Set("Cache-Control", "max-age=30")

	templinfo := getInfo(r)
	templinfo["AllRepos"] = allrepos
	err := templates.ExecuteTemplate(w, "repos.html", templinfo)
	if err != nil {
		log.Printf("template execution: %s", err)
	}
}

func servecss(w http.ResponseWriter, r *http.Request) {
	w.Header().Set("Cache-Control", "max-age=7776000")
	http.ServeFile(w, r, "views"+r.URL.Path)
}

func robotstxt(w http.ResponseWriter, r *http.Request) {
	allrepos := loadrepos()
	io.WriteString(w, "User-agent: *\n")
	for _, r := range allrepos {
		fmt.Fprintf(w, "Disallow: /r/%s/v/\n", r.Name)
	}
}

type RepoRow struct {
	ID          int
	Name        string
	Description string
	Order       int
}

func loadrepos() []*RepoRow {
	db := getdbconn()
	defer putdbconn(db)
	rows, _ := db.Conn.Query("select repoid, name, description, orderid from repos order by orderid asc")
	defer rows.Close()
	var repos []*RepoRow
	for rows.Next() {
		var repo RepoRow
		rows.Scan(&repo.ID, &repo.Name, &repo.Description, &repo.Order)
		repos = append(repos, &repo)
	}
	return repos
}

func initrepos() {
	log.Printf("init repos")
	cachemtx.Lock()
	for _, cache := range wirecaches {
		cache.Quit()
	}
	wirecaches = make(map[string]*Cache)
	cachemtx.Unlock()

	allrepos := loadrepos()

	repopoolmtx.Lock()
	for _, pool := range repogercpool {
		pool.dead = true
	}
	repogercpool = make(map[string]*ConnectionPool)
	for _, pool := range repocmdpool {
		pool.dead = true
	}
	repocmdpool = make(map[string]*ConnectionPool)
	for _, pool := range repowirepool {
		pool.dead = true
	}
	repowirepool = make(map[string]*ConnectionPool)
	for _, repo := range allrepos {
		name := repo.Name
		repogercpool[name] = newConnectionPool(name, 2, creategercconn)
		repocmdpool[name] = newConnectionPool(name, 2, createcmdconn)
		repowirepool[name] = newConnectionPool(name, 2, createwireconn)
	}
	repopoolmtx.Unlock()
}

func refresher(debug bool, updaterequestline <-chan struct{}) {
	var tickytack <-chan time.Time
	if !debug {
		tickytack = time.Tick(5 * time.Minute)
	}
	for {
		select {
		case <-tickytack:
		case <-updaterequestline:
		}
		allrepos := loadrepos()
		for _, repo := range allrepos {
			pullupdates(repo.Name)
		}
	}
}

func dontdie() {
	recover()
}

func pullupdates(name string) {
	defer dontdie()
	conn := getcmdconn(name)
	if conn == nil {
		return
	}
	defer putbackconn(conn)
	output := conn.runcmdraw("pull")
	if strings.Contains(string(output), "adding changeset") {
		conn.runcmdraw("up")

		repopoolmtx.Lock()
		repogercpool[name].dead = true
		repocmdpool[name].dead = true
		repowirepool[name].dead = true
		repogercpool[name] = newConnectionPool(name, 2, creategercconn)
		repocmdpool[name] = newConnectionPool(name, 2, createcmdconn)
		repowirepool[name] = newConnectionPool(name, 2, createwireconn)
		repopoolmtx.Unlock()

		cachemtx.Lock()
		cache := wirecaches[name]
		if cache != nil {
			cache.Flush()
		}
		cachemtx.Unlock()

		log.Printf("upped %s", name)
	}
}

func adminlistener(debug bool, updaterequestline chan<- struct{}) {
	var v struct{}
	if !debug {
		updaterequestline <- v
	}
	sockname := "humungus.sock"
	err := os.Remove(sockname)
	if err != nil && !os.IsNotExist(err) {
		log.Printf("unable to unlink admin socket: %s", err)
		return
	}
	sock, err := net.Listen("unix", sockname)
	if err != nil {
		log.Printf("unable to create admin socket: %s", err)
		return
	}
	for {
		fd, err := sock.Accept()
		if err != nil {
			continue
		}
		scanner := bufio.NewScanner(fd)
		for scanner.Scan() {
			cmd := scanner.Text()
			log.Printf("got cmd %s\n", cmd)
			switch cmd {
			case "init":
				initrepos()
				updaterequestline <- v
			case "stop":
				os.Exit(0)
			}
		}
		fd.Close()
	}
}

var savedstyleparam string

func getstyleparam() string {
	if savedstyleparam != "" {
		return savedstyleparam
	}
	data, _ := ioutil.ReadFile("views/style.css")
	hasher := sha512.New()
	hasher.Write(data)
	return fmt.Sprintf("?v=%.8x", hasher.Sum(nil))
}

func getconfig(key string, value interface{}) error {
	db := getdbconn()
	defer putdbconn(db)
	rows, err := db.Conn.Query("select value from config where key = ?", key)
	if err != nil {
		return err
	}
	defer rows.Close()
	if rows.Next() {
		err = rows.Scan(value)
	}
	return err
}

func openListener() (net.Listener, error) {
	var listenAddr string
	err := getconfig("listenaddr", &listenAddr)
	if err != nil {
		return nil, err
	}
	if listenAddr == "" {
		return nil, fmt.Errorf("must have listenaddr")
	}
	proto := "tcp"
	if listenAddr[0] == '/' {
		proto = "unix"
		err := os.Remove(listenAddr)
		if err != nil && !os.IsNotExist(err) {
			log.Printf("unable to unlink socket: %s", err)
		}
	}
	listener, err := net.Listen(proto, listenAddr)
	if err != nil {
		return nil, err
	}
	if proto == "unix" {
		os.Chmod(listenAddr, 0777)
	}
	return listener, nil
}

var preservehooks []func()

func serve() {
	listener, err := openListener()
	if err != nil {
		log.Fatal(err)
	}

	for _, h := range preservehooks {
		h()
	}

	initrepos()
	initLexers()
	hlcache = NewCache(40 * 1024 * 1024)

	debug := false
	getconfig("debug", &debug)
	templates = ParseTemplates(debug,
		"views/head.html",
		"views/path.html",
		"views/changeset.html",
		"views/changes.html",
		"views/downloads.html",
		"views/file.html",
		"views/files.html",
		"views/overview.html",
		"views/repos.html",
	)
	if !debug {
		savedstyleparam = getstyleparam()
	}

	err = getconfig("servername", &serverName)
	if err != nil {
		log.Fatalf("error querying servername: %s", err)
	}

	updaterequestline := make(chan struct{})
	go refresher(debug, updaterequestline)
	go adminlistener(debug, updaterequestline)
	log.Printf("starting web services")
	mux := mux.NewRouter()
	mux.HandleFunc("/style.css", servecss)
	mux.HandleFunc("/robots.txt", robotstxt)
	mux.HandleFunc("/", rootpage)
	mux.HandleFunc("/r/{reponame}", showrepo)
	mux.HandleFunc("/r/{reponame}/d", showdownloads)
	mux.HandleFunc("/r/{reponame}/d/{filename:[[:alnum:]_. -]*}", downloadfile)
	mux.HandleFunc("/r/{reponame}/f", showfiles)
	mux.HandleFunc("/r/{reponame}/h", showchanges)
	mux.HandleFunc("/r/{reponame}/rss", showreporss)
	mux.HandleFunc("/r/{reponame}/v/{changeset:[^/]*}", showchangeset)
	mux.HandleFunc("/r/{reponame}/v/{changeset:[^/]*}/a/{filename:[[:alnum:]_. /-]*}", showannotate)
	mux.HandleFunc("/r/{reponame}/v/{changeset:[^/]*}/d/{filename:[[:alnum:]_. /-]*}", showfileraw)
	mux.HandleFunc("/r/{reponame}/v/{changeset:[^/]*}/f/{filename:[[:alnum:]_. /-]*}", showfile)
	mux.HandleFunc("/r/{reponame}/v/{changeset:[^/]*}/h/{filename:[[:alnum:]_. /-]*}", showchanges)
	err = http.Serve(listener, mux)
	if err != nil {
		log.Fatalf("error serving: %s", err)
	}
}

func main() {
	args := os.Args[1:]

	if len(args) == 0 {
		args = []string{"serve"}
	}
	switch args[0] {
	case "admin":
		adminscreen()
	case "control":
		controlcommand(args[1:])
	case "pyggo":
		pyggo(args[1:])
	case "serve":
		serve()
	default:
		fmt.Printf("unknown command\n")
		os.Exit(1)
	}
}
