//
// Copyright (c) 2018 Ted Unangst <tedu@tedunangst.com>
//
// Permission to use, copy, modify, and distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
// ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
// OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

package main

import (
	"bytes"
	"compress/zlib"
	"crypto/sha512"
	"encoding/binary"
	"fmt"
	"io"
	"log"
	"net/http"
	"net/url"
	"regexp"
	"sort"
	"strconv"
	"strings"

	"github.com/gorilla/mux"
)

// functions to talk to an hg repo over the command server
func readdata(rpipe io.Reader) []byte {
	var numbytes int32
	var buf [1]byte
	rpipe.Read(buf[:])
	switch buf[0] {
	case 'o':
		binary.Read(rpipe, binary.BigEndian, &numbytes)
		var data []byte = make([]byte, numbytes)
		io.ReadFull(rpipe, data)
		return data
	case 'r':
		binary.Read(rpipe, binary.BigEndian, &numbytes)
		var rv int32
		binary.Read(rpipe, binary.BigEndian, &rv)
		return nil
	case 'e':
		binary.Read(rpipe, binary.BigEndian, &numbytes)
		var data []byte = make([]byte, numbytes)
		rpipe.Read(data)
		return data
	}
	var nextbytes [50]byte
	rpipe.Read(nextbytes[:])
	panic("bad news bears " + string(buf[0]) + string(nextbytes[:]))
}

func (conn *RepoConn) runcmdraw(cmd string) []byte {
	amt := int32(len(cmd))
	io.WriteString(conn.Wpipe, "runcommand\n")
	binary.Write(conn.Wpipe, binary.BigEndian, &amt)
	io.WriteString(conn.Wpipe, cmd)
	var data []byte
	for {
		somedata := readdata(conn.Rpipe)
		if somedata == nil {
			break
		}
		data = append(data, somedata...)
	}
	return data
}

func (conn *RepoConn) runcmd(cmd string, args []string) []byte {
	if args != nil {
		cmd += "\x00" + strings.Join(args, "\x00")
	}

	data := conn.runcmdraw(cmd)

	return data
}

func (conn *RepoConn) runcmdlines(cmd string, args []string) []string {
	data := conn.runcmd(cmd, args)
	lines := strings.Split(string(data), "\n")
	if len(lines) > 0 && lines[len(lines)-1] == "" {
		lines = lines[:len(lines)-1]
	}
	return lines
}

// functions to talk to an hg wire server
// mostly to support pull
func readamt(rpipe io.Reader) (int, error) {
	var onebyte [1]byte
	lenbuf := make([]byte, 0, 10)
	for {
		amt, err := rpipe.Read(onebyte[:])
		if err != nil {
			return 0, err
		}
		if amt == 0 || onebyte[0] == '\n' {
			break
		}
		lenbuf = append(lenbuf, onebyte[0])
	}
	return strconv.Atoi(string(lenbuf))
}

func pullhandler(w http.ResponseWriter, r *http.Request, reponame string, args url.Values) {
	conn := getwireconn(reponame)
	if conn == nil {
		http.NotFound(w, r)
		return
	}
	defer putbackconn(conn)

	if len(args) != 2 {
		log.Printf("cmd with missing arg")
		http.NotFound(w, r)
		return
	}

	hasher := sha512.New()
	io.WriteString(hasher, "getbundle")
	io.WriteString(hasher, "\x00")
	var keys []string
	for k, v0 := range args {
		v := v0[0]
		s := fmt.Sprintf("%s\x00%s\x00", k, v)
		keys = append(keys, s)
	}
	sort.Strings(keys)
	for _, s := range keys {
		io.WriteString(hasher, s)
	}
	key := string(hasher.Sum(nil))
	pulldata := conn.cache.Get(key)
	if pulldata != nil {
		log.Printf("cached pull for %s", reponame)
		w.Header().Set("Content-Type", "application/mercurial-0.1")
		w.Write(pulldata)
		return
	}

	log.Printf("pulling from %s", reponame)

	wpipe := conn.Wpipe
	io.WriteString(wpipe, "getbundle")
	io.WriteString(wpipe, "\n")
	io.WriteString(wpipe, fmt.Sprintf("* %d\n", len(args)))
	for k, v0 := range args {
		v := v0[0]
		fmt.Fprintf(wpipe, "%s %d\n", k, len(v))
		io.WriteString(wpipe, v)
	}
	rpipe := conn.Rpipe

	var bigbuf bytes.Buffer
	zwriter := zlib.NewWriter(&bigbuf)
	defer zwriter.Close()
	prevzero := false
	for {
		var chunklen int32
		err := binary.Read(rpipe, binary.BigEndian, &chunklen)
		if err != nil {
			log.Println("error reading bundle len", err)
			return
		}
		binary.Write(zwriter, binary.BigEndian, chunklen)
		if chunklen == 0 {
			// it seems bundle ends with two zero chunks
			if prevzero {
				break
			}
			prevzero = true
			continue
		}
		prevzero = false
		chunklen -= 4
		if chunklen > 0 {
			clen := int64(chunklen)
			_, err := io.CopyN(zwriter, rpipe, clen)
			if err != nil {
				log.Printf("problem copying chunk: %s", err)
				// it's broken, we've lost sync
				conn.Wpipe.Close()
				conn.Wpipe = nil
				return
			}
		}
	}
	zwriter.Flush()

	pulldata = bigbuf.Bytes()
	conn.cache.Put(key, pulldata)

	w.Header().Set("Content-Type", "application/mercurial-0.1")
	w.Write(pulldata)
}

func hgpassthru(w http.ResponseWriter, r *http.Request) {
	reponame := mux.Vars(r)["reponame"]

	r.ParseForm()
	cmd := r.FormValue("cmd")
	xarg := r.Header.Get("X-Hgarg-1") + r.Header.Get("X-Hgarg-2") + r.Header.Get("X-Hgarg-3")
	args, err := url.ParseQuery(xarg)
	if err != nil {
		log.Printf("failure to parse x-arg: %s: %s", xarg, err)
		http.NotFound(w, r)
		return
	}

	log.Printf("request for %s: '%s' '%s' %d", reponame, cmd, xarg, len(args))

	if cmd == "getbundle" {
		pullhandler(w, r, reponame, args)
		return
	}

	switch cmd {
	case "hello", "capabilities", "branchmap", "heads":
		if len(args) != 0 {
			log.Printf("bad capabilities cmd")
			http.NotFound(w, r)
			return
		}
	case "listkeys", "known", "lookup":
		if len(args) != 1 {
			log.Printf("cmd with missing arg")
			http.NotFound(w, r)
			return
		}
	case "batch":
		if len(args) != 1 {
			log.Printf("bad batch cmd")
			http.NotFound(w, r)
			return
		}
		subcmds := strings.Split(args.Get("cmds"), ";")
		for _, c := range subcmds {
			if c == "heads " {
			} else if strings.HasPrefix(c, "known nodes=") {
			} else {
				log.Printf("bad batch cmd %s", c)
				http.NotFound(w, r)
				return
			}
		}
	case "unbundle":
		log.Printf("push attempt")
		http.Error(w, "push denied", 403)
		return
	default:
		log.Printf("did not know that command")
		http.NotFound(w, r)
		return
	}

	conn := getwireconn(reponame)
	if conn == nil {
		http.NotFound(w, r)
		return
	}
	defer putbackconn(conn)

	wpipe := conn.Wpipe
	io.WriteString(wpipe, cmd)
	io.WriteString(wpipe, "\n")
	if len(xarg) > 0 {
		switch cmd {
		case "batch":
			io.WriteString(wpipe, "* 0\n")
		case "known":
			io.WriteString(wpipe, "* 0\n")
		}
		for k, v0 := range args {
			v := v0[0]
			fmt.Fprintf(wpipe, "%s %d\n", k, len(v))
			io.WriteString(wpipe, v)
		}
	}
	rpipe := conn.Rpipe
	w.Header().Set("Content-Type", "application/mercurial-0.1")

	datalen, _ := readamt(rpipe)
	data := make([]byte, datalen)
	io.ReadFull(rpipe, data)
	if cmd == "capabilities" {
		re := regexp.MustCompile(" bundle2=[^ ]*")
		data = re.ReplaceAll(data, []byte{})
		if bytes.Index(data, []byte("httpheader")) == -1 {
			data = append(data, []byte(" httpheader=1024")...)
		}
		datalen = len(data)
	}
	w.Header().Set("Content-Length", strconv.Itoa(datalen))
	w.Write(data)
}
