//
// Copyright (c) 2018 Ted Unangst <tedu@tedunangst.com>
//
// Permission to use, copy, modify, and distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
// ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
// OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

// something like pygments in go

package main

import (
	"bufio"
	"html/template"
	"io"
	"io/ioutil"
	"log"
	"os"
	"regexp"
	"strconv"
	"strings"
	"time"
)

type Token struct {
	name      string
	re        *regexp.Regexp
	state     int
	nextstate int
}

type Highlighter struct {
	before []byte
	after  []byte
}

func newHighlighter(before string, after string) *Highlighter {
	return &Highlighter{[]byte(before), []byte(after)}
}

var ttyhighlighters = make(map[string]*Highlighter)
var htmlhighlighters = make(map[string]*Highlighter)

func newToken(name string, regex string) *Token {
	m := strings.Split(name, ":")
	state := 0
	nextstate := 0
	if len(m) == 3 {
		name = m[0]
		state, _ = strconv.Atoi(m[1])
		nextstate, _ = strconv.Atoi(m[2])
	}
	return &Token{
		name,
		regexp.MustCompile("^" + regex),
		state,
		nextstate,
	}
}

func readTokens(file string) ([]*Token, error) {
	var tokens []*Token
	fd, err := os.Open(file)
	if err != nil {
		return tokens, err
	}
	defer fd.Close()
	scanner := bufio.NewScanner(fd)
	for scanner.Scan() {
		line := scanner.Text()
		m := strings.SplitN(line, " ", 2)
		tokens = append(tokens, newToken(m[0], m[1]))
	}
	tokens = append(tokens, newToken("unknown", "(?s:.)"))
	return tokens, nil
}

var lexers = make(map[string][]*Token)

func initLexers() {
	dirname := "lexers"
	dir, err := os.Open(dirname)
	if err != nil {
		return
	}
	defer dir.Close()

	names, _ := dir.Readdirnames(0)

	for _, name := range names {
		if !strings.HasSuffix(name, ".txt") {
			continue
		}
		tokens, err := readTokens(dirname + "/" + name)
		if err != nil {
			continue
		}
		lang := name[0 : len(name)-4]
		lexers[lang] = tokens
	}

	ttyhighlighters["keyword"] = newHighlighter("\x1b[33m", "\x1b[0m")
	ttyhighlighters["builtin"] = newHighlighter("\x1b[32m", "\x1b[0m")
	ttyhighlighters["string"] = newHighlighter("\x1b[31m", "\x1b[0m")
	ttyhighlighters["number"] = newHighlighter("\x1b[31m", "\x1b[0m")
	ttyhighlighters["type"] = newHighlighter("\x1b[32m", "\x1b[0m")
	ttyhighlighters["comment"] = newHighlighter("\x1b[34m", "\x1b[0m")
	ttyhighlighters["addline"] = newHighlighter("\x1b[32m", "\x1b[0m")
	ttyhighlighters["delline"] = newHighlighter("\x1b[31m", "\x1b[0m")

	htmlhighlighters["keyword"] = newHighlighter("<span class=kw>", "</span>")
	htmlhighlighters["builtin"] = newHighlighter("<span class=bi>", "</span>")
	htmlhighlighters["string"] = newHighlighter("<span class=st>", "</span>")
	htmlhighlighters["number"] = newHighlighter("<span class=nm>", "</span>")
	htmlhighlighters["type"] = newHighlighter("<span class=tp>", "</span>")
	htmlhighlighters["comment"] = newHighlighter("<span class=cm>", "</span>")
	htmlhighlighters["addline"] = newHighlighter("<span class=al>", "</span>")
	htmlhighlighters["delline"] = newHighlighter("<span class=dl>", "</span>")
}

func highlightcode(data []byte, filename string, w io.Writer, ishtml bool) {
	dot := strings.LastIndex(filename, ".")
	ext := filename[dot+1:]

	highlighters := ttyhighlighters

	if ishtml {
		highlighters = htmlhighlighters
	}

	tokens := lexers[ext]
	if tokens == nil {
		if ishtml {
			template.HTMLEscape(w, data)
		} else {
			w.Write(data)
		}
		return
	}

	start := time.Now()

	dataloc := 0
	state := 0
	for dataloc < len(data) {
	restart:
		for _, tok := range tokens {
			if tok.state != state {
				continue
			}
			m := tok.re.Find(data[dataloc:])
			if m != nil {
				state = tok.nextstate
				hl := highlighters[tok.name]
				if hl != nil {
					w.Write(hl.before)
				}
				if ishtml {
					template.HTMLEscape(w, m)
				} else {
					w.Write(m)
				}
				if hl != nil {
					w.Write(hl.after)
				}
				dataloc += len(m)
				goto restart
			}
		}
		state = 0
	}
	log.Printf("highlight elapsed time: %s", time.Since(start))
}

func pyggo(args []string) {
	initLexers()

	for _, fname := range args {
		data, _ := ioutil.ReadFile(fname)
		w := bufio.NewWriter(os.Stdout)
		highlightcode(data, fname, w, false)
		w.Flush()
	}
}
